/*
  Copyright 2002 Massachusetts Institute of Technology
    
  Permission to use, copy, modify, and distribute this program for any
  purpose and without fee is hereby granted, provided that this
  copyright and permission notice appear on all copies and supporting
  documentation, the name of M.I.T. not be used in advertising or
  publicity pertaining to distribution of the program without specific
  prior permission, and notice be given in supporting documentation that
  copying and distribution is by permission of M.I.T.  M.I.T. makes no
  representations about the suitability of this software for any
  purpose.  It is provided "as is" without express or implied warranty.
*/
package jsdsi;

import jsdsi.sexp.*;
import java.util.*;

/**
   An ACL entry that specifies a permission (the Tag) and a set of
   principals (the Subject) that may access the object protected by this
   entry's ACL.
   
   @see Acl
   
   @author Sameer Ajmani
 **/
public class AclEntry extends Obj {
  private final Subject subject;
  private final Auth auth;
  private final Validity validity;
  private final String comment;

  public AclEntry(Subject s, Tag t, boolean p, Validity v, String c)
  {
    assert(s != null): "null subject";
    assert(t != null): "null tag";
    subject = s;
    auth = new Auth(t, p);
    validity = v; // may be null
    comment = c;  // may be null
  }

  public Subject getSubject()
  {
    return subject;
  }

  public Tag getTag()
  {
    return auth.getTag();
  }

  public boolean getPropagate()
  {
    return auth.getPropagate();
  }

  public Validity getValidity()
  {
    return validity;
  }

  public String getComment()
  {
    return comment;
  }

  public boolean equals(Object o)
  {
    if (o instanceof AclEntry) {
      AclEntry e = (AclEntry)o;
      return subject.equals(e.subject)
        && auth.equals(e.auth)
        && Util.equals(validity, e.validity)
        && Util.equals(comment, e.comment);
    }
    return false;
  }

  public int hashCode()
  {
    return subject.hashCode()
      ^ auth.hashCode()
      ^ Util.hashCode(validity)
      ^ Util.hashCode(comment);
  }

  public SexpList toSexp()
  {
    List l = new ArrayList(5);
    l.add(getSubject().toSexp());
    if (getPropagate()) {
      l.add(SexpUtil.toSexp("propagate"));
    }
    l.add(getTag().toSexp());
    if (getValidity() != null) {
      l.add(getValidity().toSexp());
    }
    if (getComment() != null) {
      l.add(SexpUtil.toSexpComment(getComment()));
    }
    return SexpUtil.toSexp("entry", l);
  }

  static AclEntry parseAclEntry(SexpList l)
    throws SexpParseException
  {
    Iterator ebody = SexpUtil.getBody(l);
    // FIXME: same as Cert parsing!
    // <sub-obj>
    Subject subject = Subject.Default.parseSubject
      (SexpUtil.getNextList(ebody, "acl entry subject"));
    // <deleg>? <tag>
    boolean propagate = false;
    SexpList propOrTag = SexpUtil.getNextList(ebody, "cert propagate or tag");
    String type = propOrTag.getType();
    if (type.equals("propagate")) {
      propagate = true;
      SexpUtil.check(propOrTag.size() == 1, "extra fields in propagate");
      propOrTag = SexpUtil.getNextList(ebody, "tag", "cert tag");
    }
    Tag tag = Tag.parseTag(propOrTag);
    // <valid>? <comment>?
    Validity validity = null;
    String comment = null;
    if (ebody.hasNext()) {
      SexpList validOrComment =
        SexpUtil.getNextList(ebody, "cert valid or comment");
      type = validOrComment.getType();
      if (type.equals("valid")) {
        validity = Validity.parseValidity(validOrComment);
        if (ebody.hasNext()) {
          validOrComment =
            SexpUtil.getNextList(ebody, "comment", "cert comment");
          type = "comment"; // FIXME: ugly!
        }
      }
      if (type.equals("comment")) {
        Iterator combody = SexpUtil.getBody(validOrComment);
        comment = SexpUtil.getNextString(combody, "comment body");
        SexpUtil.checkDone(combody, "comment");
      }
    }
    return new AclEntry(subject, tag, propagate, validity, comment);
  }
}
