/*
  Copyright 2002 Massachusetts Institute of Technology
    
  Permission to use, copy, modify, and distribute this program for any
  purpose and without fee is hereby granted, provided that this
  copyright and permission notice appear on all copies and supporting
  documentation, the name of M.I.T. not be used in advertising or
  publicity pertaining to distribution of the program without specific
  prior permission, and notice be given in supporting documentation that
  copying and distribution is by permission of M.I.T.  M.I.T. makes no
  representations about the suitability of this software for any
  purpose.  It is provided "as is" without express or implied warranty.
*/
package jsdsi;

import jsdsi.sexp.*;
import java.util.*;

/**
   A SPKI authorization certificate: conveys a permission (the Tag) from
   the issuer to the subject.  If the propagate flag is set, the subject
   can further delegate the permission.

   @see Cert
   
   @author Sameer Ajmani
 **/
public class AuthCert extends Cert {
  private final Auth auth;

  public AuthCert(Principal i, Subject s, Validity v,
                  String d, String c, Tag t, boolean p)
  {
    super(i, s, v, d, c);
    assert(t != null): "null tag";
    auth = new Auth(t, p);
  }

  public Tag getTag() {
    return auth.getTag();
  }

  public boolean getPropagate() {
    return auth.getPropagate();
  }

  public boolean equals(Object o) {
    if (o instanceof AuthCert) {
      AuthCert a = (AuthCert)o;
      return auth.equals(a.auth)
        && super.equals(o);
    }
    return false;
  }

  public int hashCode() {
    return auth.hashCode()
      ^ super.hashCode();
  }

  public SexpList toSexp()
  {
    List l = new ArrayList(5);

    // issuer block
    Sexp[] is = new Sexp[1];
    is[0] = getIssuer().toSexp();
    l.add(SexpUtil.toSexp("issuer", is));

    // subject block
    Sexp[] ss = new Sexp[1];
    // FIXME: can we eliminate these downcasts?
    if (getSubject() instanceof Name) {
      // tell name about issuer
      ss[0] = ((Name)getSubject()).toSexp(getIssuer());
    } else if (getSubject() instanceof Threshold) {
      // tell nested names about issuer
      ss[0] = ((Threshold)getSubject()).toSexp(getIssuer());      
    } else {
      ss[0] = getSubject().toSexp();
    }
    l.add(SexpUtil.toSexp("subject", ss));

    // auth block
    if (getPropagate()) {
      l.add(SexpUtil.toSexpList("propagate"));
    }
    l.add(getTag().toSexp());
    
    if (getValidity() != null) {
      l.add(getValidity().toSexp());
    }
    if (getComment() != null) {
      l.add(SexpUtil.toSexpComment(getComment()));
    }
    return SexpUtil.toSexp("cert", l);
  }
}

