/*
  Copyright 2002 Massachusetts Institute of Technology
    
  Permission to use, copy, modify, and distribute this program for any
  purpose and without fee is hereby granted, provided that this
  copyright and permission notice appear on all copies and supporting
  documentation, the name of M.I.T. not be used in advertising or
  publicity pertaining to distribution of the program without specific
  prior permission, and notice be given in supporting documentation that
  copying and distribution is by permission of M.I.T.  M.I.T. makes no
  representations about the suitability of this software for any
  purpose.  It is provided "as is" without express or implied warranty.
*/
package jsdsi;

import jsdsi.sexp.*;
import java.util.*;

/**
   A single SPKI/SDSI certificate.

   @see AuthCert
   @see NameCert
   
   @author Sameer Ajmani
 **/
public abstract class Cert extends Obj implements Element {
  private final Principal issuer;
  private final Subject subject;
  private final Validity validity;
  private final String display;
  private final String comment;
  
  public Cert(Principal i, Subject s, Validity v, String d, String c)
  {
    assert(i != null): "null issuer";
    assert(s != null): "null subject";
    issuer = i;
    subject = s;
    validity = v; // may be null
    display = d;  // may be null
    comment = c;  // may be null
  }
  
  public Principal getIssuer() {
    return issuer;
  }
  
  public Subject getSubject() {
    return subject;
  }

  public Validity getValidity() {
    return validity;
  }

  public String getComment() {
    return comment;
  }

  public String getDisplay() {
    return display;
  }

  public boolean equals(Object o) {
    if (o instanceof Cert) {
      Cert c = (Cert)o;
      return issuer.equals(c.issuer)
        && subject.equals(c.subject)
        && Util.equals(validity, c.validity)
        && Util.equals(display, c.display)
        && Util.equals(comment, c.comment);
    }
    return false;
  }

  public int hashCode() {
    return issuer.hashCode()
      ^ subject.hashCode()
      ^ Util.hashCode(validity)
      ^ Util.hashCode(display)
      ^ Util.hashCode(comment);
  }

  static Cert parseCert(SexpList l)
    throws SexpParseException
  {
    Iterator cbody = SexpUtil.getBody(l);
    Iterator ibody = SexpUtil.getBody
      (SexpUtil.getNextList(cbody, "issuer", "cert issuer"));
    SexpList nameOrPrincipal = SexpUtil.getNextList(ibody, "issuer body");
    SexpUtil.checkDone(ibody, "issuer");
    String type = nameOrPrincipal.getType();
    Principal issuer = null;
    String name = null;
    if (type.equals("name")) {
      Iterator nbody = SexpUtil.getBody(nameOrPrincipal);
      issuer = Principal.parsePrincipal
        (SexpUtil.getNextList(nbody, "name issuer"));
      name = SexpUtil.getNextString(nbody, "name string");
      SexpUtil.checkDone(nbody, "issuer-name");
    } else {
      // FIXME: defaulting to principal is confusing
      issuer = Principal.parsePrincipal(nameOrPrincipal);
    }

    // (subject <subj-obj>)
    Iterator sbody = SexpUtil.getBody
      (SexpUtil.getNextList(cbody, "subject", "cert subject"));
    Subject subject = Subject.Default.parseSubject
      (SexpUtil.getNextList(sbody, "subject body"), issuer);
    SexpUtil.checkDone(sbody, "subject");
    
    Tag tag = null;
    boolean propagate = false;
    if (name == null) {
      // <deleg>? <tag>
      SexpList propOrTag = SexpUtil.getNextList
        (cbody, "cert propagate or tag");
      type = propOrTag.getType();
      if (type.equals("propagate")) {
        propagate = true;
        SexpUtil.check(propOrTag.size() == 1, "extra fields in propagate");
        propOrTag = SexpUtil.getNextList(cbody, "tag", "cert tag");
      }
      tag = Tag.parseTag(propOrTag);
    }
    // <valid>? <comment>?
    Validity validity = null;
    String comment = null;
    if (cbody.hasNext()) {
      SexpList validOrComment = SexpUtil.getNextList
        (cbody, "cert valid or comment");
      type = validOrComment.getType();
      if (type.equals("valid")) {
        validity = Validity.parseValidity(validOrComment);
        if (cbody.hasNext()) {
          validOrComment = SexpUtil.getNextList
            (cbody, "comment", "cert comment");
          type = "comment"; // FIXME: ugly!
        }
      }
      if (type.equals("comment")) {
        Iterator combody = SexpUtil.getBody(validOrComment);
        comment = SexpUtil.getNextString(combody, "comment body");
        SexpUtil.checkDone(combody, "comment");
      }
    }
    SexpUtil.checkDone(cbody, "cert");
    SexpUtil.check((name == null) != (tag == null), // sanity check
                   "internal error: either name or tag must not be null");
    if (name == null) {
      return new AuthCert(issuer, subject, validity, null, // TODO: display
                          comment, tag, propagate);
    } else {
      return new NameCert(issuer, subject, validity, null, // TODO: display
                          comment, name);
    }
  }
}

