/*
  Copyright 2002 Massachusetts Institute of Technology
    
  Permission to use, copy, modify, and distribute this program for any
  purpose and without fee is hereby granted, provided that this
  copyright and permission notice appear on all copies and supporting
  documentation, the name of M.I.T. not be used in advertising or
  publicity pertaining to distribution of the program without specific
  prior permission, and notice be given in supporting documentation that
  copying and distribution is by permission of M.I.T.  M.I.T. makes no
  representations about the suitability of this software for any
  purpose.  It is provided "as is" without express or implied warranty.
*/
package jsdsi;

import java.security.*;
import java.security.cert.*;
import java.util.*;

/**
   Attempts to create a certification path that satisfies the given
   parameters: essentially a wrapper around the Prover class and its
   subclasses.

   @see CertPathParameters
   @see Prover
   
   @author Sameer Ajmani
 **/
public class CertPathBuilder extends CertPathBuilderSpi
{  
  public java.security.cert.CertPathBuilderResult
    engineBuild(java.security.cert.CertPathParameters params)
    throws CertPathBuilderException,
           InvalidAlgorithmParameterException
  {
    try {
      return engineBuild((jsdsi.CertPathParameters)params);
    } catch (ClassCastException e) {
      throw (InvalidAlgorithmParameterException)
        new InvalidAlgorithmParameterException().initCause(e);
    }
  }
  
  public jsdsi.CertPathBuilderResult
    engineBuild(jsdsi.CertPathParameters params)
    throws CertPathBuilderException,
           InvalidAlgorithmParameterException
  {
    if (params instanceof IssuerCertPathParameters) {
      return engineBuild(new FProver(params.getCert(), params.getStore()));
    }
    if (params instanceof SubjectCertPathParameters) {
      return engineBuild(new RProver(params.getCert(), params.getStore()));
    }
    throw new InvalidAlgorithmParameterException
      ("unrecognized params: "+params.getClass().getName());
  }

  private jsdsi.CertPathBuilderResult
    engineBuild(Prover p)
    throws CertPathBuilderException
  {
      Proof pf = p.getProof();
      if (pf == null) {
        throw new CertPathBuilderException("no proof found");
      }
      return new CertPathBuilderResult
        (new CertPath(pf),
         new CertPathBuilderStats(p.getNumFetched()));
  }

  public static void main(String[] args)
    throws Exception
  {
    Provider.install();
    if (args.length != 2) {
      System.err.println
        ("usage: java CertPathBuilder <forward?> <input-file>");
      return;
    }
    boolean forward = args[0].equals("true");
    Loader l = new Loader(args[1]);
    java.security.cert.CertPathBuilder builder =
      java.security.cert.CertPathBuilder.getInstance("SPKI");

    // try to find proofs from name to each key
    Iterator ni = l.getNames().iterator();
    while (ni.hasNext()) {
      Name n = (Name)ni.next();
      Iterator ki = l.getKeys().iterator();
      while (ki.hasNext()) {
        PublicKey k = (PublicKey)ki.next();
        NameCert c = new NameCert
          (n.getIssuer(), k, null, null, null, n.getNames()[0]);
        try {
          Proof pf = buildProof(builder, c, l.getCertStore(), forward);
          NameCert nc = (NameCert)pf.getCert();
          System.out.println("RESULT: Proof for "
                             +whois(nc.getIssuer())
                             +" "+nc.getName()+" -> "
                             +whois((Principal)nc.getSubject()));
          System.out.println(pf);
        } catch (CertPathBuilderException e) {
          System.out.println("RESULT: No proof for "
                             +n.getNames()[0]+" -> "+whois(k));
        }
      }
    }

    // try to find proofs with each tag, with and without propagate
    Iterator ii = l.getKeys().iterator();
    while (ii.hasNext()) {
      PublicKey issuer = (PublicKey)ii.next();
      Iterator ki = l.getKeys().iterator();
      while (ki.hasNext()) {
        PublicKey k = (PublicKey)ki.next();
        Iterator ti = l.getTags().iterator();
        while (ti.hasNext()) {
          Tag t = (Tag)ti.next();
          for (int i = 0; i < 2; i++) {
            boolean propagate = (i == 0);
            AuthCert c = new AuthCert
              (issuer, k, null, null, null, t, propagate);
            try {
              Proof pf = buildProof(builder, c, l.getCertStore(), forward);
              AuthCert ac = (AuthCert)pf.getCert();
              System.out.println("RESULT: Proof for "
                                 +whois(ac.getIssuer())
                                 +" "+(ac.getPropagate()?"+":"!")
                                 +((StringTag)ac.getTag()).getValue()+" -> "
                                 +whois((Principal)ac.getSubject()));
              System.out.println(pf);
            } catch (CertPathBuilderException e) {
              System.out.println("RESULT: No proof for "
                                 +whois(issuer)
                                 +" "+(propagate?"+":"!")
                                 +((StringTag)t).getValue()+" -> "
                                 +whois(k));
            }
          }
        }
      }
    }
  }
  
  private static String whois(Principal p)
  {
    return new String(((RSAPublicKey)p).getModulus().toByteArray());
  }
  
  private static Proof
    buildProof(java.security.cert.CertPathBuilder builder, Cert c,
               java.security.cert.CertStore s, boolean forward)
    throws CertPathBuilderException
  {
    try {
      jsdsi.CertPathParameters params;
      if (forward) {
        params = new IssuerCertPathParameters(c, s);
      } else {
        params = new SubjectCertPathParameters(c, s);
      }
      jsdsi.CertPathBuilderResult res =
        (jsdsi.CertPathBuilderResult)builder.build(params);
      System.out.println("num fetched = "+res.getStats().getNumFetched());
      return ((jsdsi.CertPath)res.getCertPath()).getProof();
    } catch (InvalidAlgorithmParameterException e) {
      throw new Error(e);
    }
  }
}
