/*
  Copyright 2002 Massachusetts Institute of Technology
    
  Permission to use, copy, modify, and distribute this program for any
  purpose and without fee is hereby granted, provided that this
  copyright and permission notice appear on all copies and supporting
  documentation, the name of M.I.T. not be used in advertising or
  publicity pertaining to distribution of the program without specific
  prior permission, and notice be given in supporting documentation that
  copying and distribution is by permission of M.I.T.  M.I.T. makes no
  representations about the suitability of this software for any
  purpose.  It is provided "as is" without express or implied warranty.
*/
package jsdsi;

import java.security.*;
import java.security.cert.*;
import java.util.*;

/**
   A Cert along with its validators (verification path, signature, and
   online test results).  Whereas a Cert is simply an unauthenticated
   statement, a Certificate is self-validating and thus can be
   considered authentic if verify() succeeds.  A Certificate is
   serialized as a SPKI/SDSI Sequence.
   
   @see Cert
   @see Signature
   @see Sequence
   
   @author Sameer Ajmani
 **/
public class Certificate extends java.security.cert.Certificate
{
  private PublicKey issuer; // if null, cert.getIssuer() is a key
  private Cert cert;
  // TODO: VerificationPath ...
  private Signature sig;
  // TODO: OnlineTestResult[] ...

  public Certificate(PublicKey k, Cert c, Signature s)
    throws CertificateException
  {
    super("SPKI");
    assert(c != null): "null cert";
    assert(s != null): "null signature";
    issuer = k;
    cert = c;
    sig = s;
    if ((issuer == null) && !(c.getIssuer() instanceof PublicKey)) {
      throw new CertificateException("issuer must be a PublicKey");
    }
  }
  
  public Certificate(Cert c, Signature s)
    throws CertificateException
  {
    this(null, c, s);
  }

  public java.security.PublicKey getPublicKey()
  {
    if (issuer == null) {
      return (PublicKey)cert.getIssuer();
    } else {
      return issuer;
    }
  }

  public void
    verify(java.security.PublicKey key)
    throws CertificateException, NoSuchAlgorithmException,
           InvalidKeyException, NoSuchProviderException,
           SignatureException
  {
    verify(key, null);
  }

  public void
    verify(java.security.PublicKey key, String sigProvider)
    throws CertificateException, NoSuchAlgorithmException,
           InvalidKeyException, NoSuchProviderException,
           SignatureException
  {
    assert(key != null): "null key";    
    if (!cert.getIssuer().samePrincipalAs(sig.getSigner())) {
      throw new CertificateException("issuer does not match signer");      
    }
    if (!(key instanceof Principal)
        || !cert.getIssuer().samePrincipalAs((Principal)key)) {
      throw new CertificateException
        ("verification key does not match issuer");
    }
    if (!sig.verify(key, cert, sigProvider)) {
      throw new SignatureException("signature verification failed");
    }
    // TODO: check verification path, online test results, etc.
  }
  
  public Cert getCert()
  {
    return cert;
  }

  private static void check(Iterator elems)
    throws CertificateException
  {
    if (!elems.hasNext()) {
      throw new CertificateException
        ("Not enough elements in sequence");
    }
  }

  public static Certificate fromElements(Iterator elems)
    throws CertificateException
  {
    PublicKey issuer = null;
    check(elems);
    Element e = (Element)elems.next();
    if (e instanceof PublicKey) {
      issuer = (PublicKey)e;
      check(elems);
      e = (Element)elems.next();
    }
    if (!(e instanceof Cert)) {
      throw new CertificateException("Expected cert");
    }
    Cert cert = (Cert)e;
    check(elems);
    e = (Element)elems.next();
    if (!(e instanceof Signature)) {
      throw new CertificateException("Expected signature");
    }
    Signature sig = (Signature)e;
    // TODO: include verification path, online test results, etc.
    return new Certificate(issuer, cert, sig);
  }

  public void toElements(List elems)
  {
    if (issuer != null) {
      elems.add(issuer);
      // TODO: if key is hashed in cert or sig, include hash ops
    }
    elems.add(cert);
    elems.add(sig);
    // TODO: include verification path, online test results, etc.
  }
  
  public static Certificate fromSequence(Sequence seq)
    throws CertificateException
  {
    return fromElements(Arrays.asList(seq.getElements()).iterator());
  }

  public Sequence toSequence()
  {
    ArrayList elems = new ArrayList();
    toElements(elems);
    return new Sequence((Element[])elems.toArray(new Element[0]));
  }

  public String toString()
  {
    return toSequence().toString();
  }

  public byte[] getEncoded()
  {
    return toSequence().toByteArray();
  }

  public String getFormat()
  {
    return "SEXP";
  }
}
