/*
  Copyright 2002 Massachusetts Institute of Technology
    
  Permission to use, copy, modify, and distribute this program for any
  purpose and without fee is hereby granted, provided that this
  copyright and permission notice appear on all copies and supporting
  documentation, the name of M.I.T. not be used in advertising or
  publicity pertaining to distribution of the program without specific
  prior permission, and notice be given in supporting documentation that
  copying and distribution is by permission of M.I.T.  M.I.T. makes no
  representations about the suitability of this software for any
  purpose.  It is provided "as is" without express or implied warranty.
*/
package jsdsi;

import jsdsi.sexp.*;
import java.net.URL;
import java.security.*;
import java.util.*;

/**
   A cryptographic hash value.  Specifies the hash algorithm, the hash
   value, and an optional set of URIs that specify where the hashed
   object is located.
   
   @author Sameer Ajmani
 **/
public class Hash extends Obj
{
  private final String algo;
  private final byte[] data;
  private final URL[] urls; // may be null

  public Hash(String a, byte[] d, URL[] u)
  {
    assert(a != null): "null algo";
    assert(d != null): "null data";
    algo = a;
    data = d;
    urls = u;
  }

  public Hash(String a, Obj o, URL[] u)
  {
    assert(a != null): "null algo";
    assert(o != null): "null object";
    algo = a;
    try {
      data = MessageDigest.getInstance(a).digest(o.toByteArray());
    } catch (NoSuchAlgorithmException e) {
      throw new Error(e);
    }
    urls = u;
  }

  public Hash(String a, byte[] d)
  {
    this(a, d, null);
  }

  public Hash(String a, Obj o)
  {
    this(a, o, null);
  }

  public String getAlgorithm()
  {
    return algo;
  }

  public byte[] getData()
  {
    return data;
  }

  public URL[] getURLs()
  {
    return urls;
  }

  public boolean equals(Object o)
  {
    if (o instanceof Hash) {
      Hash h = (Hash)o;
      return algo.equals(h.algo)
        && java.util.Arrays.equals(data, h.data)
        && Util.equals(urls, h.urls);
    }
    return false;
  }

  public int hashCode()
  {
    return algo.hashCode()
      ^ data.hashCode()
      ^ Util.hashCode(urls);
  }

  public SexpList toSexp()
  {
    List l = new ArrayList(3);
    l.add(SexpUtil.toSexp(getAlgorithm()));
    l.add(SexpUtil.toSexp(getData()));
    if (getURLs() != null) {
      l.add(SexpUtil.toSexp(getURLs()));
    }
    return SexpUtil.toSexp("hash", l);
  }
  
  public static Hash parseHash(SexpList l)
    throws SexpParseException
  {
    Iterator hbody = SexpUtil.getBody(l);
    String algo = SexpUtil.getNextString(hbody, "hash algo");
    byte[] data = SexpUtil.getNextByteArray(hbody, "hash data");
    SexpUtil.checkDone(hbody, "hash"); // TODO: support URIs
    return new Hash(algo, data);
  }
}
