/*
  Copyright 2002 Massachusetts Institute of Technology
    
  Permission to use, copy, modify, and distribute this program for any
  purpose and without fee is hereby granted, provided that this
  copyright and permission notice appear on all copies and supporting
  documentation, the name of M.I.T. not be used in advertising or
  publicity pertaining to distribution of the program without specific
  prior permission, and notice be given in supporting documentation that
  copying and distribution is by permission of M.I.T.  M.I.T. makes no
  representations about the suitability of this software for any
  purpose.  It is provided "as is" without express or implied warranty.
*/
package jsdsi;

import java.io.*;
import java.util.*;
import java.math.*;
import java.security.*;
import java.security.cert.*;

/**
   <p>Creates a set of certificates from a flat text file for testing.
   Each certificate appears on its own line.  Blank lines are allowed,
   but comments are not.

   <p>Name certs are specified as follows:
   <br>ISSUER name -> SUBJECT [names...]

   <p>For example,
   <br>ALICE friends -> BOB
   <br>BOB my-friends -> BOB sister
   <br>BOB sister -> CAROL

   <p>Note that we use uppercase for keys and lowecase for names for
   clarity, but this is not required.  However, lines are
   case-sensitive, so "BOB" is different from "Bob".

   <p>Auth certs are specified as follows:
   <br>ISSUER [!|+]tag -> SUBJECT [names...]
   <br>where ! means the permission "tag" cannot be delegated, while "+"
   means that it can.

   <p>For example,
   <br>ALICE !read -> BOB
   <br>BOB +write -> BOB my-friends   
   
   @author Sameer Ajmani
 **/
public class Loader {
  private Map keys = new HashMap();
  private Set certSet = new HashSet();
  private Set nameSet = new HashSet();
  private Set tagSet = new HashSet();
  private java.security.cert.CertStore store;
  public Loader(String filename) throws IOException
  {
    Set certs = new HashSet();
    LineNumberReader in = new LineNumberReader(new FileReader(filename));
    in.setLineNumber(1);
    for (String line = in.readLine(); line != null; line = in.readLine()) {
      StringTokenizer t = new StringTokenizer(line);
      if (t.countTokens() == 0) {
        continue; // skip empty lines
      }
      if (t.countTokens() < 4) {
        throw new IOException("bad input on line "
                              +in.getLineNumber()+": "+line);
      }
      Principal issuer = getPrincipal(t.nextToken());
      String name = t.nextToken();
      String arrow = t.nextToken();
      if (!arrow.equals("->")) {
        throw new IOException("bad arrow on line "
                              +in.getLineNumber()+": "+line);
      }
      Principal sub = getPrincipal(t.nextToken());
      String[] names = new String[t.countTokens()];
      for (int i = 0; i < names.length; i++) {
        names[i] = t.nextToken();
      }
      Subject subject = (names.length > 0)
        ? (Subject)new Name(sub, names) : (Subject)sub;

      // create a fake certificate with a fake signature
      Cert c;
      if (name.startsWith("!") || name.startsWith("+")) {
        // this is an auth cert
        Tag tag = new StringTag(name.substring(1));
        tagSet.add(tag);
        c = new AuthCert(issuer, subject, null, null, null,
                         tag, name.startsWith("+")); // propagate?
      } else {
        // this is a name cert
        nameSet.add(new Name(issuer, name));
        c = new NameCert(issuer, subject, null, null, null, name);
      }
      certSet.add(c);
      Signature s = new Signature
        (issuer, new Hash("md5", "HASH-VALUE".getBytes(), null),
         "rsa-pkcs1-md5", "SIGNATURE-VALUE".getBytes());
      try {
        certs.add(new Certificate(c, s));
      } catch (java.security.cert.CertificateException e) {
        throw new Error(e);
      }
    }
    try {
      store = java.security.cert.CertStore.getInstance
        ("SPKI", new CollectionCertStoreParameters(certs));
    } catch (InvalidAlgorithmParameterException e) {
      throw new Error(e);
    } catch (NoSuchAlgorithmException e) {
      throw new Error(e);
    }
  }
  public java.security.cert.CertStore getCertStore()
  {
    return store;
  }
  public Collection getKeys()
  {
    return keys.values();
  }
  public Set getCerts()
  {
    return certSet;
  }
  public Set getNames()
  {
    return nameSet;
  }
  public Set getTags()
  {
    return tagSet;
  }
  private Principal getPrincipal(String id)
  {
    RSAPublicKey k = (RSAPublicKey)keys.get(id);
    if (k == null) {
      k = new RSAPublicKey(new BigInteger(id.getBytes()),
                           new BigInteger(new byte[] {0x03}),
                           "MD5/RSA/PKCS#1", null);
      keys.put(id, k);
    }
    return k;
  }
}
