/*
  Copyright 2002 Massachusetts Institute of Technology
    
  Permission to use, copy, modify, and distribute this program for any
  purpose and without fee is hereby granted, provided that this
  copyright and permission notice appear on all copies and supporting
  documentation, the name of M.I.T. not be used in advertising or
  publicity pertaining to distribution of the program without specific
  prior permission, and notice be given in supporting documentation that
  copying and distribution is by permission of M.I.T.  M.I.T. makes no
  representations about the suitability of this software for any
  purpose.  It is provided "as is" without express or implied warranty.
*/
package jsdsi;

import jsdsi.sexp.*;
import java.util.*;

/**
   A SDSI name: a principal and a sequence of strings.  If the sequence
   contains just one string, this is a local name; otherwise this is an
   extended name.
   
   @author Sameer Ajmani
 **/
public class Name extends Obj implements Subject
{
  private final Principal issuer;
  private final String[] names;

  public Name(Principal i, String[] n)
  {
    assert(i != null): "null issuer";
    assert(n != null): "null names";
    assert(n.length > 0) : "empty names";
    issuer = i;
    names = n;
  }

  public Name(Principal i, String n)
  {
    this(i, new String[] { n }); // XXX assert n != null
  }

  public Principal getIssuer()
  {
    return issuer;
  }

  public String[] getNames()
  {
    return names;
  }

  public boolean prefixOf(Name n)
  {
    if (!issuer.equals(n.issuer)) {
      return false;
    }
    if (names.length > n.names.length) {
      return false;
    }
    for (int i = 0; i < names.length; i++) {
      if (!names[i].equals(n.names[i])) {
        return false;
      }
    }
    return true;
  }

  public Name prefix(int i)
  {
    assert(i <= names.length) : "prefix too long";
    String[] ns = new String[i];
    System.arraycopy(names, 0, ns, 0, i);
    return new Name(issuer, ns);
  }

  public Name prefix()
  {
    return prefix(1);
  }

  public boolean equals(Object o)
  {
    if (o instanceof Name) {
      Name n = (Name)o;
      return issuer.equals(n.issuer)
        && Util.equals(names, n.names);
    }
    return false;
  }

  public int hashCode()
  {
    return issuer.hashCode() ^ Util.hashCode(names);
  }

  public SexpList toSexp()
  {
    return toSexp((Principal)null);
  }
  
  public SexpList toSexp(Principal iss)
  {
    List l = new ArrayList();
    if (!getIssuer().samePrincipalAs(iss)) {
      // make fully-qualified name
      l.add(getIssuer().toSexp());
    }
    for (int i = 0; i < names.length; i++) {
      l.add(SexpUtil.toSexp(names[i]));
    }
    return SexpUtil.toSexp("name", l);
  }

  static Name parseName(SexpList l)
    throws SexpParseException
  {
    return parseName(l, null);
  }
  
  static Name parseName(SexpList l, Principal issuerParam)
    throws SexpParseException
  {
    Iterator nbody = SexpUtil.getBody(l);
    Sexp s = SexpUtil.getNext(nbody, "first name component");
    Principal issuer = null;
    String[] names = null;
    if (s instanceof SexpList) {
      // name is fully-qualified
      issuer = Principal.parsePrincipal(SexpUtil.getList(s));
      names = new String[l.size()-2];      
      for (int i = 0; i < names.length; i++) {
        names[i] = SexpUtil.getNextString(nbody, "name string #"+i);
      }
    } else if (issuerParam != null) {
      // name is relative; issuer provided as a parameter
      issuer = issuerParam;
      names = new String[l.size()-1];
      names[0] = SexpUtil.getString(s);
      for (int i = 1; i < names.length; i++) {
        names[i] = SexpUtil.getNextString(nbody, "name string #"+i);
      }
    } else {
      throw new SexpParseException
        ("relative name used without issuer");
    }
    SexpUtil.checkDone(nbody, "name"); // sanity check
    return new Name(issuer, names);    
  }

  // TODO: boolean sameNameAs(Name n)
  // TODO: boolean sameNameAs(Principal i, String n)
}

