/*
  Copyright 2002 Massachusetts Institute of Technology
    
  Permission to use, copy, modify, and distribute this program for any
  purpose and without fee is hereby granted, provided that this
  copyright and permission notice appear on all copies and supporting
  documentation, the name of M.I.T. not be used in advertising or
  publicity pertaining to distribution of the program without specific
  prior permission, and notice be given in supporting documentation that
  copying and distribution is by permission of M.I.T.  M.I.T. makes no
  representations about the suitability of this software for any
  purpose.  It is provided "as is" without express or implied warranty.
*/
package jsdsi;

import jsdsi.sexp.*;
import java.security.*;
import java.security.cert.*;
import java.util.*;

/**
   A statement (a Cert) and a sequence of Certificates (Certs +
   validators) that proves that the statement holds.  Proofs are
   self-validating and can be composed to create new proofs.
   
   @see Certificate
   @see Cert
   
   @author Sameer Ajmani
 **/
public class Proof extends Obj
{
  public static class IncompatibleException extends Exception {
    IncompatibleException(String message)
    {
      super(message);
    }
  }
  private Cert cert;         // the statement to prove
  private Certificate[] certs; // the certificates that prove it
  public Proof(Certificate c)
  {
    assert(c != null): "null certificate";
    cert = c.getCert();
    certs = new Certificate[] { c };
  }
  private Proof(Cert c, Certificate[] cs)
  {
    assert(c != null): "null cert";
    assert(cs != null): "null certificates";
    cert = c;
    certs = cs;
  }
  public Cert getCert() {
    return cert;
  }
  public Certificate[] getCertificates() {
    return certs;
  }
  public Sequence getSequence() {
    List elems = new ArrayList();
    for (int i = 0; i < certs.length; i++) {
      certs[i].toElements(elems);
    }
    Element[] es = new Element[elems.size()];
    return new Sequence((Element[])elems.toArray(es));
  }
  public boolean equals(Object o) {
    if (o instanceof Proof) {
      Proof p = (Proof)o;
      return cert.equals(p.cert)
        && Util.equals(certs, p.certs);
    }
    return false;
  }
  public int hashCode() {
    return cert.hashCode() ^ Util.hashCode(certs);
  }

  public SexpList toSexp()
  {
    Sexp[] ss = new Sexp[2];
    ss[0] = getCert().toSexp();
    ss[1] = getSequence().toSexp();
    return SexpUtil.toSexp("proof", ss);
  }

  public void verify()
    throws CertificateException, NoSuchAlgorithmException,
           InvalidKeyException, NoSuchProviderException,
           SignatureException
  {
    if (certs.length == 0) {
      throw new CertificateException("no certificates");
    }
    certs[0].verify(certs[0].getPublicKey());
    Proof p = new Proof(certs[0]);
    try {
      for (int i = 1; i < certs.length; i++) {
        certs[i].verify(certs[i].getPublicKey());
        p = p.compose(new Proof(certs[i]));
      }
    } catch (IncompatibleException e) {
      throw (CertificateException)new CertificateException
        ("certificates do not compose correctly").initCause(e);      
    }
    if (!p.getCert().equals(cert)) {
      throw new CertificateException
        ("composed certificates do not match proof");
    }
  }
  private Subject composeNames(Name lhName, Name rhName, Subject rhSubject)
    throws IncompatibleException
  {
    if (!rhName.prefixOf(lhName)) {
      throw new IncompatibleException
        ("Cannot compose "+lhName+" with "+rhName);        
    }
    int diff = lhName.getNames().length - rhName.getNames().length;
    if (diff == 0) {
      return rhSubject;
    }
    assert(diff > 0): "expected righthand name length <= lefthand";
    if (!(rhSubject instanceof Principal)) {
      throw new IncompatibleException
        ("Cannot compose name with non-reducing cert");      
    }
    // replace rhName with rhSubject in lhName
    String[] names = new String[diff];
    int off = lhName.getNames().length - diff;
    for (int i = 0; i < names.length; i++) {
      names[i] = lhName.getNames()[off+i];
    }
    return new Name((Principal)rhSubject, names);
  }
  private Subject composeWithNameCert(Cert lhs, NameCert rhs)
    throws IncompatibleException
  {
    if (!(lhs.getSubject() instanceof Name)) {
      throw new IncompatibleException
        ("Cannot compose reducing Cert with NameCert");
    }
    Name lhName = (Name)lhs.getSubject();
    // TODO: allow variable-length names on RHS
    Name rhName = rhs.getFullName();
    return composeNames(lhName, rhName, rhs.getSubject());
  }
  private Validity intersect(Validity v1, Validity v2) {
    if (v1 == null) {
      return v2;
    }
    if (v2 == null) {
      return v1;
    }
    return v1.intersect(v2);
  }
  private NameCert composeNameName(NameCert lhs, NameCert rhs)
    throws IncompatibleException
  {
    return new NameCert
      (lhs.getIssuer(),
       composeWithNameCert(lhs, rhs),
       intersect(lhs.getValidity(), rhs.getValidity()),
       lhs.getDisplay(),
       lhs.getComment(),
       lhs.getName());
  }  
  private AuthCert composeAuthName(AuthCert lhs, NameCert rhs)
    throws IncompatibleException
  {
    return new AuthCert
      (lhs.getIssuer(),
       composeWithNameCert(lhs, rhs),
       intersect(lhs.getValidity(), rhs.getValidity()),
       lhs.getDisplay(),
       lhs.getComment(),
       lhs.getTag(),
       lhs.getPropagate());
  }
  private AuthCert composeAuthAuth(AuthCert lhs, AuthCert rhs)
    throws IncompatibleException
  {
    if (!lhs.getPropagate()) {
      throw new IncompatibleException("cannot propagate this auth");
    }
    return new AuthCert
      (lhs.getIssuer(),
       rhs.getSubject(),
       intersect(lhs.getValidity(), rhs.getValidity()),
       lhs.getDisplay(),
       lhs.getComment(),
       lhs.getTag().intersect(rhs.getTag()),
       rhs.getPropagate());
  }
  private Certificate[] concat(Proof p)
  {
    Certificate[] cs = new Certificate[certs.length + p.certs.length];
    System.arraycopy(certs, 0, cs, 0, certs.length);
    System.arraycopy(p.certs, 0, cs, certs.length, p.certs.length);
    return cs;
  }
  /**
     Composes this proof with another proof.
     
     @param p the proof to compose this with
     @return a new proof that is the composition of this with p.
     @throws IncompatibleException if this cannot be composed with p.
  **/
  public Proof compose(Proof p) throws IncompatibleException
  {
    if (cert instanceof NameCert) {
      if (!(p.cert instanceof NameCert)) {
        throw new IncompatibleException
          ("Cannot compose NameCert with: "+p.cert.getClass().getName());
      }
      return new Proof
        (composeNameName((NameCert)cert, (NameCert)p.cert),
         concat(p));
    }
    if (cert instanceof AuthCert) {
      if (p.cert instanceof NameCert) {
        return new Proof
          (composeAuthName((AuthCert)cert, (NameCert)p.cert),
           concat(p));
      }
      if (p.cert instanceof AuthCert) {
        return new Proof
          (composeAuthAuth((AuthCert)cert, (AuthCert)p.cert),
           concat(p));
      }
    }
    throw new ClassCastException
      ("Unrecognized Cert subclass: "+cert.getClass().getName());
  }
}
