/*
  Copyright 2002 Massachusetts Institute of Technology
    
  Permission to use, copy, modify, and distribute this program for any
  purpose and without fee is hereby granted, provided that this
  copyright and permission notice appear on all copies and supporting
  documentation, the name of M.I.T. not be used in advertising or
  publicity pertaining to distribution of the program without specific
  prior permission, and notice be given in supporting documentation that
  copying and distribution is by permission of M.I.T.  M.I.T. makes no
  representations about the suitability of this software for any
  purpose.  It is provided "as is" without express or implied warranty.
*/
package jsdsi;

import java.security.*;
import java.security.cert.*;
import java.util.*;

/**
   Given a statement (a Cert) and a CertStore, attempts to construct a
   Proof that the statement holds using certificates from the store.
   
   @author Sameer Ajmani
 **/
abstract class Prover
{
  static class ProofFoundException extends Exception
  {
    Proof proof;
    ProofFoundException(Proof p) {
      super("found proof: "+p);
      proof = p;
    }
    Proof getProof()
    {
      return proof;
    }
  }
  Cert provee; // the statement to prove
  java.security.cert.CertStore store;
  MultiMap check = new MultiMap(); // cert -> set of proof(cert)
  MultiMap value = new MultiMap(); // name -> set of proof(name -> principal)
  MultiMap compatible = new MultiMap(); // name -> set of proof(LHS -> name+X)
  MultiMap issuer = new MultiMap(); // issuer -> set of proof(issuer -> RHS)
  MultiMap reverse = new MultiMap(); // subject -> set of proof(LHS -> subject)

  // statistics
  private int numFetched = 0; // number of certs fetched from cert store
  int getNumFetched()
  {
    return numFetched;
  }

  Prover(Cert c, java.security.cert.CertStore s)
  {
    provee = c;
    store = s;
  }

  private boolean attempted = false;
  private Proof proof; // the proof; null if no proof found
  public final Proof getProof()
  {
    if (!attempted) {
      attempted = true;
      proof = makeProof();
    }
    return proof;
  }
  
  abstract Proof makeProof();
  abstract void insert(Proof p) throws ProofFoundException;

  void insertCertificates(Collection certs)
    throws ProofFoundException
  {
    // insert stored proofs
    //System.out.println("INSERT: inserting certs "+certs.size());
    Iterator i = certs.iterator();
    while (i.hasNext()) {
      insert(new Proof((Certificate)i.next()));
    }
  }
  
  Set load(Set cache, Object key, CertSelector sel, MultiMap map)
    throws ProofFoundException
  {
    try {
      if (!cache.contains(key)) {
        cache.add(key);
        // fetch stored proofs
        try {
          Collection stored = store.getCertificates(sel);
          numFetched += stored.size();
          insertCertificates(stored);
        } catch (CertStoreException e) {
          throw new Error(e);
        }
      }
      return map.get(key);
    } catch (ProofFoundException e) {
      // invalidate cache
      cache.remove(key);
      throw e;
    }
  }
}
