/*
  Copyright 2002 Massachusetts Institute of Technology
    
  Permission to use, copy, modify, and distribute this program for any
  purpose and without fee is hereby granted, provided that this
  copyright and permission notice appear on all copies and supporting
  documentation, the name of M.I.T. not be used in advertising or
  publicity pertaining to distribution of the program without specific
  prior permission, and notice be given in supporting documentation that
  copying and distribution is by permission of M.I.T.  M.I.T. makes no
  representations about the suitability of this software for any
  purpose.  It is provided "as is" without express or implied warranty.
*/
package jsdsi;

import jsdsi.sexp.*;
import java.util.*;
import java.net.URL;

/**
   A public key.
   
   @author Sameer Ajmani
 **/
public abstract class PublicKey extends Principal
  implements Element, java.security.PublicKey
{
  private final String algo;
  private final URL[] urls;

  public PublicKey(String a, URL[] u)
  {
    assert(a != null): "null algo";
    algo = a;
    urls = u; // may be null
  }

  public PublicKey(String a)
  {
    this(a, null);
  }

  public String getAlgorithm()
  {
    return algo;
  }

  public String getFormat()
  {
    return "SEXP";
  }

  public byte[] getEncoded()
  {
    return toByteArray();
  }

  public URL[] getURLs()
  {
    return urls;
  }

  public boolean samePrincipalAs(Principal p)
  {
    if (p == null) {
      return false;
    }
    if (p instanceof PublicKey) {
      return equals(p);
    }
    // p is a PublicKeyHash
    return p.samePrincipalAs(this);
  }

  public boolean equals(Object o)
  {
    if (o instanceof PublicKey) {
      PublicKey p = (PublicKey)o;
      return algo.equals(p.algo)
        && Util.equals(urls, p.urls);
    }
    return false;
  }

  public int hashCode()
  {
    return algo.hashCode() ^ Util.hashCode(urls);
  }

  abstract SexpList toPublicKeySexp();
  public final SexpList toSexp()
  {
    List l = new ArrayList(2);
    l.add(toPublicKeySexp());
    if (getURLs() != null) {
      l.add(SexpUtil.toSexp(getURLs()));
    }
    return SexpUtil.toSexp("public-key", l);
  }

  static PublicKey parsePublicKey(SexpList l)
    throws SexpParseException
  {
    Iterator kbody = SexpUtil.getBody(l);
    SexpList key = SexpUtil.getNextList(kbody, "public key body");
    SexpUtil.checkDone(kbody, "public-key"); // TODO: support URIs
    String type = key.getType();
    if (type.startsWith("rsa")) {
      return RSAPublicKey.parseRSAPublicKey(key, "rsa");
    }
    throw new SexpParseException("unrecognized public-key type: "+type);    
  }
}
