/*
  Copyright 2002 Massachusetts Institute of Technology
    
  Permission to use, copy, modify, and distribute this program for any
  purpose and without fee is hereby granted, provided that this
  copyright and permission notice appear on all copies and supporting
  documentation, the name of M.I.T. not be used in advertising or
  publicity pertaining to distribution of the program without specific
  prior permission, and notice be given in supporting documentation that
  copying and distribution is by permission of M.I.T.  M.I.T. makes no
  representations about the suitability of this software for any
  purpose.  It is provided "as is" without express or implied warranty.
*/
package jsdsi;

import jsdsi.sexp.*;
import java.math.BigInteger;
import java.net.URL;
import java.security.*;
import java.util.*;

/**
   An RSA public key.
   
   @author Sameer Ajmani
 **/
public class RSAPublicKey extends PublicKey
  implements java.security.interfaces.RSAPublicKey
{
  private final BigInteger modulus;
  private final BigInteger exponent;

  public RSAPublicKey(BigInteger m, BigInteger e, String a, URL[] u)
  {
    super(a, u);
    assert(m != null): "null modulus";
    assert(e != null): "null exponent";
    modulus = m;
    exponent = e;
  }

  public RSAPublicKey(BigInteger m, BigInteger e, String a)
  {
    this(m, e, a, null);
  }

  public RSAPublicKey(java.security.interfaces.RSAPublicKey k, URL[] u)
  {
    this(k.getModulus(), k.getPublicExponent(), k.getAlgorithm(), u);
  }

  public RSAPublicKey(java.security.interfaces.RSAPublicKey k)
  {
    this(k, null);
  }

  /**
     Creates a new RSA key pair whose public key is a Principal.
     @param a the specific RSA algorithm to use
     @param provider the provider to use
   **/
  public static KeyPair create(String a, String provider)
    throws NoSuchAlgorithmException, NoSuchProviderException
  {
    assert(a != null): "null algo";
    KeyPairGenerator kpg = (provider == null)
      ? KeyPairGenerator.getInstance(a)
      : KeyPairGenerator.getInstance(a, provider);
    KeyPair kp = kpg.genKeyPair();
    assert(kp.getPublic() instanceof java.security.interfaces.RSAPublicKey)
      : "did not generate RSA key";
    return new KeyPair
      (new RSAPublicKey((java.security.interfaces.RSAPublicKey)kp.getPublic()),
       kp.getPrivate());
  }

  /**
     Creates a new RSA key pair whose public key is a Principal.
     @param a the specific RSA algorithm to use
   **/
  public static KeyPair create(String a)
    throws NoSuchAlgorithmException, NoSuchProviderException
  {
    return create(a, null);
  }
  
  /**
     Creates a new RSA key pair whose public key is a Principal.
   **/
  public static KeyPair create()
    throws NoSuchAlgorithmException, NoSuchProviderException
  {
    return create("RSA");
  }
  
  public BigInteger getModulus() {
    return modulus;
  }

  public BigInteger getExponent() {
    return exponent;
  }

  public BigInteger getPublicExponent() {
    return exponent;
  }
  
  public boolean equals(Object o) {
    if (o instanceof RSAPublicKey) {
      RSAPublicKey r = (RSAPublicKey)o;
      return modulus.equals(r.modulus)
        && exponent.equals(r.exponent)
        && super.equals(o);
    }
    return false;
  }

  public int hashCode() {
    return modulus.hashCode()
      ^ exponent.hashCode()
      ^ super.hashCode();
  }

  public SexpList toPublicKeySexp()
  {
    String algo = "rsa";
    Sexp[] ss = new Sexp[2];
    Sexp[] es = new Sexp[1];
    es[0] = SexpUtil.toSexp(getExponent().toByteArray());
    ss[0] = SexpUtil.toSexp("e", es);
    Sexp[] ns = new Sexp[1];
    ns[0] = SexpUtil.toSexp(getModulus().toByteArray());
    ss[1] = SexpUtil.toSexp("n", ns);
    return SexpUtil.toSexp(algo, ss);
  }
  
  static RSAPublicKey parseRSAPublicKey(SexpList l, String algo)
    throws SexpParseException
  {
    Iterator rbody = SexpUtil.getBody(l);
    Iterator ebody = SexpUtil.getBody
      (SexpUtil.getNextList(rbody, "e", "exponent"));
    BigInteger e = new BigInteger
      (SexpUtil.getNextByteArray(ebody, "e value"));
    SexpUtil.checkDone(ebody, "exponent");
    Iterator nbody = SexpUtil.getBody
      (SexpUtil.getNextList(rbody, "n", "modulus"));
    BigInteger n = new BigInteger
      (SexpUtil.getNextByteArray(nbody, "n value"));
    SexpUtil.checkDone(nbody, "modulus");
    SexpUtil.checkDone(rbody, "rsa public-key");
    return new RSAPublicKey(n, e, algo);
  }
}
