/*
  Copyright 2002 Massachusetts Institute of Technology
    
  Permission to use, copy, modify, and distribute this program for any
  purpose and without fee is hereby granted, provided that this
  copyright and permission notice appear on all copies and supporting
  documentation, the name of M.I.T. not be used in advertising or
  publicity pertaining to distribution of the program without specific
  prior permission, and notice be given in supporting documentation that
  copying and distribution is by permission of M.I.T.  M.I.T. makes no
  representations about the suitability of this software for any
  purpose.  It is provided "as is" without express or implied warranty.
*/
package jsdsi;

import jsdsi.sexp.*;
import java.util.*;

/**
   A tag that specifies a range of allowed values.
   
   @author Sameer Ajmani
 **/
public class RangeTag extends ExprTag
{
  private final String ordering;
  private final String upperLimit;
  private final boolean strictUpper;
  private final String lowerLimit;
  private final boolean strictLower;

  public RangeTag(String o, String u, boolean su, String l, boolean sl)
  {
    assert(o != null): "null ordering";
    assert(u != null): "null upper bound";
    assert(l != null): "null lower bound";
    ordering = o;
    upperLimit = u;
    strictUpper = su;
    lowerLimit = l;
    strictLower = sl;
  }

  boolean implies(Tag that)
  {
    return false; // TODO: implement
  }
  
  public boolean equals(Object that)
  {
    if (that instanceof RangeTag) {
      RangeTag r = (RangeTag)that;
      return this.ordering.equals(r.ordering)
        && this.upperLimit.equals(r.upperLimit)
        && this.strictUpper == r.strictUpper
        && this.lowerLimit.equals(r.lowerLimit)
        && this.strictLower == r.strictLower;
    }
    return false;
  }

  public int hashCode()
  {
    return ordering.hashCode()
      ^ upperLimit.hashCode()
      ^ lowerLimit.hashCode()
      ^ (strictUpper ? 1 : 0)
      ^ (strictLower ? 1 : 0);
  }

  public String getOrdering()
  {
    return ordering;
  }

  public String getUpperLimit()
  {
    return upperLimit;
  }

  public String getLowerLimit()
  {
    return lowerLimit;
  }

  public boolean upperIsStrict()
  {
    return strictUpper;
  }

  public boolean lowerIsStrict()
  {
    return strictLower;
  }

  public Sexp toTagSexp()
  {
    Sexp[] ss = new Sexp[6];
    ss[0] = SexpUtil.toSexp("range");
    ss[1] = SexpUtil.toSexp(getOrdering());
    ss[2] = SexpUtil.toSexp(lowerIsStrict()?"g":"ge");
    ss[3] = SexpUtil.toSexp(getLowerLimit());
    ss[4] = SexpUtil.toSexp(upperIsStrict()?"l":"le");
    ss[5] = SexpUtil.toSexp(getUpperLimit());
    return SexpUtil.toSexp("*", ss);
  }

  static RangeTag parseRangeTag(Iterator tbody)
    throws SexpParseException
  {
    String ordering = SexpUtil.getNextString(tbody, "range ordering");
    // lower strictness and limit
    String gte = SexpUtil.getNextString(tbody, "range lower strictness");
    boolean strictLower = gte.equals("g");
    if (!strictLower && !gte.equals("ge"))
      throw new SexpParseException
        ("range lower strictness is not 'g' or 'ge': "+gte);
    String lowerLimit = SexpUtil.getNextString(tbody, "range lower limit");
    // upper strictness and limit
    String lte = SexpUtil.getNextString(tbody, "range gte");
    boolean strictUpper = lte.equals("l");
    if (!strictUpper && !lte.equals("le"))
      throw new SexpParseException
        ("range upper strictness is not 'l' or 'le': "+lte);
    String upperLimit = SexpUtil.getNextString(tbody, "range upper limit");
    SexpUtil.checkDone(tbody, "range tag");
    return new RangeTag(ordering,
                        upperLimit, strictUpper,
                        lowerLimit, strictLower);
  }
}

