/*
  Copyright 2002 Massachusetts Institute of Technology
    
  Permission to use, copy, modify, and distribute this program for any
  purpose and without fee is hereby granted, provided that this
  copyright and permission notice appear on all copies and supporting
  documentation, the name of M.I.T. not be used in advertising or
  publicity pertaining to distribution of the program without specific
  prior permission, and notice be given in supporting documentation that
  copying and distribution is by permission of M.I.T.  M.I.T. makes no
  representations about the suitability of this software for any
  purpose.  It is provided "as is" without express or implied warranty.
*/
package jsdsi;

import jsdsi.sexp.*;
import java.security.*;
import java.util.*;

/**
   A cryptographic signature.
   
   @author Sameer Ajmani
 **/
public class Signature extends Obj implements Element
{
  private final Principal signer;
  private final Hash signee;
  private final String algo;
  private final byte[] data;

  public Signature(Principal i, Hash h, String a, byte[] d)
  {
    assert(i != null): "null signer";
    assert(h != null): "null signee";
    assert(a != null): "null algo";
    assert(d != null): "null data";
    signer = i;
    signee = h;
    algo = a;
    data = d;
  }

  public static Signature create(java.security.KeyPair kp,
                          Obj o, String a)
    throws NoSuchAlgorithmException, InvalidKeyException,
           NoSuchProviderException, SignatureException
  {
    return create(kp, o, a, null);
  }

  public static Signature create(java.security.KeyPair kp,
                          Obj o, String a, String provider)
    throws NoSuchAlgorithmException, InvalidKeyException,
           NoSuchProviderException, SignatureException
  {
    assert(kp != null): "null key pair";
    assert(o != null): "null signee object";
    assert(a != null): "null algo";
    assert(kp.getPublic() instanceof Principal)
      : "public key must be a principal";
    java.security.Signature sig = (provider == null)
      ? java.security.Signature.getInstance(a)
      : java.security.Signature.getInstance(a, provider);
    sig.initSign(kp.getPrivate());
    sig.update(o.toByteArray());
    byte[] d = sig.sign();
    return new Signature((Principal)kp.getPublic(),
                         new Hash("md5", o),
                         a, d);
  }

  public boolean
    verify(java.security.PublicKey key, Obj o)
    throws NoSuchAlgorithmException, InvalidKeyException,
           NoSuchProviderException, SignatureException
  {
    return verify(key, o, null);
  }

  public boolean
    verify(java.security.PublicKey key, Obj o, String provider)
    throws NoSuchAlgorithmException, InvalidKeyException,
           NoSuchProviderException, SignatureException
  {
    java.security.Signature sig = (provider == null)
      ? java.security.Signature.getInstance(algo)
      : java.security.Signature.getInstance(algo, provider);
    sig.initVerify(key);
    sig.update(o.toByteArray());
    return sig.verify(data);
  }

  public Principal getSigner() {
    return signer;
  }

  public Hash getSignee() {
    return signee;
  }

  public String getAlgorithm() {
    return algo;
  }

  public byte[] getData() {
    return data;
  }

  public boolean equals(Object o) {
    if (o instanceof Signature) {
      Signature s = (Signature)o;
      return signer.equals(s.signer)
        && signee.equals(s.signee)
        && algo.equals(s.algo)
        && Util.equals(data, s.data);
    }
    return false;
  }

  public int hashCode() {
    return signer.hashCode()
      ^ signee.hashCode()
      ^ algo.hashCode()
      ^ Util.hashCode(data);
  }

  public SexpList toSexp()
  {
    Sexp[] ss = new Sexp[3];
    ss[0] = getSignee().toSexp();
    ss[1] = getSigner().toSexp();
    Sexp[] sig = new Sexp[1];
    sig[0] = SexpUtil.toSexp(getData());
    ss[2] = SexpUtil.toSexp(getAlgorithm(), sig);
    return SexpUtil.toSexp("signature", ss);
  }

  static Signature parseSignature(SexpList l)
    throws SexpParseException
  {
    Iterator sbody = SexpUtil.getBody(l);
    // (signature <hash> <principal> (<algo> <data))
    Hash signee = Hash.parseHash(SexpUtil.getNextList(sbody, "sig hash"));
    Principal signer = Principal.parsePrincipal
      (SexpUtil.getNextList(sbody, "sig principal"));
    SexpList algoblock = SexpUtil.getNextList(sbody, "sig algo block");
    SexpUtil.checkDone(sbody, "signature");
    String algo = algoblock.getType();
    Iterator algobody = SexpUtil.getBody(algoblock);
    byte[] data = SexpUtil.getNextByteArray(algobody, "sig data");
    SexpUtil.checkDone(algobody, "signature");
    return new Signature(signer, signee, algo, data);
  }
}
