/*
  Copyright 2002 Massachusetts Institute of Technology
    
  Permission to use, copy, modify, and distribute this program for any
  purpose and without fee is hereby granted, provided that this
  copyright and permission notice appear on all copies and supporting
  documentation, the name of M.I.T. not be used in advertising or
  publicity pertaining to distribution of the program without specific
  prior permission, and notice be given in supporting documentation that
  copying and distribution is by permission of M.I.T.  M.I.T. makes no
  representations about the suitability of this software for any
  purpose.  It is provided "as is" without express or implied warranty.
*/
package jsdsi;

import jsdsi.sexp.*;
import java.util.*;

/**
   A validity period and a set of online tests.
   
   @author Sameer Ajmani
 **/
public class Validity extends Obj
{
  private final Date notBefore; // may be null
  private final Date notAfter; // may be null
  private final OnlineTest[] tests;
  // TODO: replace with Set

  public Validity(Date b, Date a, OnlineTest[] t)
  {
    notBefore = b;
    notAfter = a;
    tests = t;
  }

  public Validity(Date b, Date a)
  {
    this(b, a, new OnlineTest[0]);
  }
  
  public boolean valid()
  {
    // FIXME: needs to run online tests (parameters?)
    Date now = new Date();
    return (notBefore == null || notBefore.before(now))
      && (notAfter == null || notAfter.after(now));
  }

  public Date getNotBefore()
  {
    return notBefore;
  }

  public Date getNotAfter()
  {
    return notAfter;
  }

  public OnlineTest[] getOnlineTests()
  {
    return tests;
  }

  public Validity intersect(Validity v)
  {
    Date nb = notBefore;
    if ((nb == null) || (v.notBefore != null && v.notBefore.after(nb))) {
      nb = v.notBefore;
    }
    Date na = notAfter;
    if ((na == null) || (v.notAfter != null && v.notAfter.before(na))) {
      na = v.notAfter;
    }
    // combine lists of tests
    OnlineTest[] ts = new OnlineTest[tests.length + v.tests.length];
    System.arraycopy(tests, 0, ts, 0, tests.length);
    System.arraycopy(v.tests, 0, ts, tests.length, v.tests.length);
    return new Validity(nb, na, ts);
  }

  public boolean equals(Object o)
  {
    if (o instanceof Validity) {
      Validity v = (Validity)o;
      return Util.equals(notBefore, v.notBefore)
        && Util.equals(notAfter, v.notAfter)
        && Util.equals(tests, v.tests);
    }
    return false;
  }

  public int hashCode()
  {
    return Util.hashCode(notBefore)
      ^ Util.hashCode(notAfter)
      ^ Util.hashCode(tests);
  }

  public SexpList toSexp()
  {
    List l = new ArrayList(2);
    if (getNotBefore() != null) {
      Sexp[] ss = new Sexp[1];
      ss[0] = SexpUtil.toSexp(getNotBefore());
      l.add(SexpUtil.toSexp("not-before", ss));
    }
    if (getNotAfter() != null) {
      Sexp[] ss = new Sexp[1];
      ss[0] = SexpUtil.toSexp(getNotAfter());
      l.add(SexpUtil.toSexp("not-after", ss));
    }
    for (int i = 0; i < tests.length; i++) {
      l.add(tests[i].toSexp());
    }
    return SexpUtil.toSexp("valid", l); 
  }
  
  static Validity parseValidity(SexpList l)
    throws SexpParseException
  {
    Iterator vbody = SexpUtil.getBody(l);
    SexpList check = SexpUtil.getNextList(vbody, "validity check");
    String type = check.getType();
    Date notBefore = null;
    Date notAfter = null;
    int numOnlineTests = l.size()-1;
    if (type.equals("not-before")) {
      numOnlineTests--;
      Iterator nb = SexpUtil.getBody(check);
      notBefore = SexpUtil.parseDate
        (SexpUtil.getNextString(nb, "not-before date"));
      SexpUtil.checkDone(nb, "not-before");
      if (vbody.hasNext()) {
        check = SexpUtil.getNextList(vbody, "validity check");
        type = check.getType();
      }
    }
    if (type.equals("not-after")) {
      numOnlineTests--;
      Iterator na = SexpUtil.getBody(check);
      notAfter = SexpUtil.parseDate
        (SexpUtil.getNextString(na, "not-after date"));
      SexpUtil.checkDone(na, "not-after");
      if (vbody.hasNext()) {
        check = SexpUtil.getNextList(vbody, "validity check");
        type = check.getType();
      }
    }
    OnlineTest[] tests = new OnlineTest[numOnlineTests];
    for (int i = 0; i < tests.length; i++) {
      tests[i] = OnlineTest.parseOnlineTest
        (SexpUtil.getNextList(vbody, "online test "+(i+1)+"/"+tests.length));
    }
    SexpUtil.checkDone(vbody, "valid");
    return new Validity(notBefore, notAfter, tests);
  }
}

